/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "gnu-pop3d.h"

static struct option long_options[] = {
  {"debug", 2, 0, 'D'},
  {"daemon", 2, 0, 'd'},
  {"group", 1, 0, 'g'},
  {"help", 0, 0, 'h'},
  {"inetd", 0, 0, 'i'},
  {"ipv6", 0, 0, '6'},
  {"mdn", 0, 0, 'm'},
  {"port", 1, 0, 'p'},
  {"timeout", 1, 0, 't'},
  {"tmppath", 1, 0, 'X'},
  {"user", 1, 0, 'u'},
  {"version", 0, 0, 'v'},
  {0, 0, 0, 0}
};

int
main (int argc, char **argv)
{
  struct group *gr;
  static int mode = INTERACTIVE;
  int option_index = 0;
  int c = 0;
  char *group_name = NULL;
  char *uid_name = NULL;
  struct passwd *temp_pw;
  char *opts = "D::d::hi6p:t:vmX:u:g:";
  char logbuf[128];

  last_signal = 0;				/* primarily for use in the client */
  exit_flag = 0;				/* parent process will exit if this becomes non-zero */
  child_procs = 0;				/* no. of child processes */
  total_sessions = 0;			/* no. of total connections */
  maxchildren = 10;
  daemonpid = getpid ();
  debug = 0;					/* no debugging by default */
  port = 110;					/* Default POP3 port */
  timeout = 600;				/* Timeout according to RFC 1939 */

  uid = 0;						/* Not superuser -- just unset */
  mdn = 0;						/* No delivery status notification by default */
  msgstatus = 1;				/* Status: headers supported and updated by default */
  mdncount = 0;
  ipv6enabled = 0;
  useroptions = NULL;
  tmppath = "/tmp";

  /* Set the signal handlers */
  memset (&signalaction, 0, sizeof (struct sigaction));
  signalaction.sa_sigaction = &pop3_signal;
  signalaction.sa_flags = SA_SIGINFO | SA_NOCLDSTOP /*| SA_RESTART */ ;
  sigemptyset (&signalaction.sa_mask);
  sigaddset (&signalaction.sa_mask, SIGHUP);
  sigaddset (&signalaction.sa_mask, SIGINT);
  sigaddset (&signalaction.sa_mask, SIGQUIT);
  sigaddset (&signalaction.sa_mask, SIGILL);
  sigaddset (&signalaction.sa_mask, SIGBUS);
  sigaddset (&signalaction.sa_mask, SIGFPE);
  sigaddset (&signalaction.sa_mask, SIGSEGV);
  sigaddset (&signalaction.sa_mask, SIGTERM);
  sigaddset (&signalaction.sa_mask, SIGPIPE);
  sigaddset (&signalaction.sa_mask, SIGCHLD);
  sigaddset (&signalaction.sa_mask, SIGUSR1);
  sigaddset (&signalaction.sa_mask, SIGUSR2);

  sigaction (SIGHUP, &signalaction, NULL);
  sigaction (SIGINT, &signalaction, NULL);
  sigaction (SIGQUIT, &signalaction, NULL);
  sigaction (SIGILL, &signalaction, NULL);
  sigaction (SIGBUS, &signalaction, NULL);
  sigaction (SIGFPE, &signalaction, NULL);
  sigaction (SIGSEGV, &signalaction, NULL);
  sigaction (SIGTERM, &signalaction, NULL);
  sigaction (SIGPIPE, &signalaction, NULL);
  sigaction (SIGUSR1, &signalaction, NULL);
  sigaction (SIGUSR2, &signalaction, NULL);

/*
  signal (SIGHUP, pop3_signal);
  signal (SIGINT, pop3_signal);
  signal (SIGQUIT, pop3_signal);
  signal (SIGILL, pop3_signal);
  signal (SIGBUS, pop3_signal);
  signal (SIGFPE, pop3_signal);
  signal (SIGSEGV, pop3_signal);
  signal (SIGTERM, pop3_signal);
*/

/* Don't die when a process goes away unexpectedly.
   Ignore write on a pipe with no reader. ? */
/*  signal (SIGPIPE, pop3_signal); */

  /* Set up for syslog */
  openlog ("gnu-pop3d", LOG_PID, LOG_MAIL);
  pop3_syslog (LOG_INFO, NULL, FALSE);	/* just initialize table */

  while ((c = getopt_long (argc, argv, opts, long_options, &option_index)) && c != -1)
	{
	  switch (c)
		{
		case 'D':
		  debug = optarg ? strtol (optarg, NULL, 16) : 0x11;
		  sprintf (logbuf, "debug=0x%.3x  STDERR:%.1x child:%.1x daemon:%.1x ", debug, debug >> 8 & 0xf,
				   debug >> 4 & 0xf, debug & 0xf);
		  pop3_syslog (LOG_INFO, logbuf, FALSE);
		  break;
		case 'd':
		  mode = DAEMON;
		  maxchildren = optarg ? atoi (optarg) : 10;
		  syslog (LOG_INFO, "maxchildren=%d", maxchildren);
		  childpids = calloc (maxchildren, sizeof (pid_t));
		  break;
		case 'g':
		  if ((group_name = (char *) malloc (strlen (optarg) + 1)) == (char *) NULL)
			{
			  fprintf (stderr, "Exiting: can't allocate memory");
			  exit (1);
			}
		  (void) strncpy (group_name, optarg, strlen (optarg) + 1);
		  break;
		case 'h':
		  pop3_usage (argv[0]);
		  exit (0);
		case 'i':
		  mode = INTERACTIVE;
		  break;
		case 'p':
		  mode = DAEMON;
		  port = atoi (optarg);
		  break;
		case '6':
#ifdef USE_IPV6
		  ipv6enabled = 1;
#else
		  ipv6enabled = 0;
		  pop3_syslog (LOG_WARNING,
					   "IPv6 support is not compiled into this version. Use configure with --enable-ipv6", FALSE);
#endif
		  break;
		case 't':
		  timeout = atoi (optarg);
		  /* RFC 1939 says no less than 10 minutes, but we'll let them use any */
		  break;
		case 'u':
		  if ((uid_name = (char *) malloc (strlen (optarg) + 1)) == (char *) NULL)
			{
			  fprintf (stderr, "Exiting: can't allocate memory");
			  exit (1);
			}
		  (void) strncpy (uid_name, optarg, strlen (optarg) + 1);
		  break;
		case 'X':
		  tmppath = optarg;
		  break;
		case 'm':				/* Message Disposition Notification */
		  mdn = 1;
		  break;
		case 'v':
		  printf ("GNU POP3 Server Version " VERSION "\n");
		  exit (0);
		  break;
		default:
		  pop3_usage (argv[0]);
		  exit (1);
		}
	}

  /* Use command-line group option or use "mail" group */
  if (!group_name)
	{
	  if ((group_name = (char *) malloc (5)) == (char *) NULL)
		{
		  fprintf (stderr, "Exiting: can't allocate memory");
		  exit (1);
		}
	  snprintf (group_name, 5, "mail");
	}

  if ((gr = getgrnam (group_name)) == NULL)
	{
	  /* then see if it is a GID */
	  gid = (gid_t) atoi (group_name);
	  /* zero is also error */
	  if ((gid_t) gid == 0)
		{
		  fprintf (stderr, "Group %s not found\n", group_name);
		  exit (1);
		}
	  if ((gr = getgrgid ((gid_t) gid)) == NULL)
		{
		  fprintf (stderr, "GID %s not found\n", group_name);
		  exit (1);
		}
	}
  else
	gid = gr->gr_gid;

  if (setgid (gr->gr_gid) == -1)
	{
	  fprintf (stderr, "Error setting group: %s (%d)\n", group_name, gr->gr_gid);
	  exit (1);
	}

  if (group_name)
	free (group_name);

  /* Use command-line user option or use "mail" user */
  if (!uid_name)
	{
	  if ((uid_name = (char *) malloc (5)) == (char *) NULL)
		{
		  fprintf (stderr, "Exiting: can't allocate memory");
		  exit (1);
		}
	  snprintf (uid_name, 5, "mail");
	}
  if (uid_name)
	{
	  if ((temp_pw = getpwnam (uid_name)) == NULL)
		{
		  /* then see if it is a UID */
		  uid = (uid_t) atoi (uid_name);
		  /* zero is also error */
		  if ((uid_t) uid == 0)
			{
			  fprintf (stderr, "User %s not found\n", uid_name);
			  exit (1);
			}
		  if ((temp_pw = getpwuid ((uid_t) uid)) == NULL)
			{
			  fprintf (stderr, "UID %s not found\n", uid_name);
			  exit (1);
			}
		}
	  uid = temp_pw->pw_uid;
	  free (uid_name);
	  /* pop3_freepw(temp_pw); weird allocation causes signal 11 */
	}
  if (mode == DAEMON)
	{
	  if (debug & 0x01)
		pop3_syslog (LOG_INFO, "before daemon_init", FALSE);

	  pop3_daemon_init ();
	  if (debug & 0x01)
		pop3_syslog (LOG_INFO, "after daemon_init", FALSE);
	}

  /* change directories */
#ifdef MAILSPOOLHOME
  chdir ("/");
#else
  chdir (_PATH_MAILDIR);
#endif

  umask (S_IROTH | S_IWOTH | S_IXOTH);	/* 007 */
  sock = -1;
  /* Actually run the daemon */
  switch (mode)
	{
	case DAEMON:
	  sprintf (logbuf, "GNU-pop3d " VERSION " started as DAEMON using port %d with %s",
			   port, (ipv6enabled == 1 ? "both IPv4 and IPv6" : "only IPv4"));
	  pop3_syslog (LOG_INFO, logbuf, TRUE);
	  pop3_daemon (maxchildren, argv);
	  break;
	case INTERACTIVE:
	default:
	  sprintf (logbuf, "%s", "GNU-pop3d " VERSION " started in INTERACTIVE mode");
	  pop3_syslog (LOG_INFO, logbuf, TRUE);
	  pop3_mainloop (fileno (stdin), fileno (stdout), argv);
	  break;
	}

  /* Close the syslog connection and exit */
  sprintf (logbuf, "GNU-pop3d terminated after %ld session(s)", total_sessions);
  pop3_syslog (LOG_INFO, logbuf, TRUE);
  closelog ();
  return OK;
}

/* Sets things up for daemon mode */

void
pop3_daemon_init (void)
{
  pid_t pgroup;
  char logbuf[128];

#ifdef FULL_SETSID_NEEDED
/* waitpid in pop3_signal is waiting for any child process, but kill(0,..) needs all processes to be in the
   same process group.
   A service started from event.d does not handle the fail safe method below, as the parent process dies */
  if (fork ())
	exit (0);					/* parent exits */
  pgroup = setsid ();			/* become session leader */

  if (fork ())
	exit (0);					/* new parent exits */
#else
  pgroup = setsid ();			/* become session leader */
#endif
  if (debug & 0x01 && pgroup < 0)
	{
	  sprintf (logbuf,
			   "pop3_daemon_init: did not set new session leader errno=%d, will use current: %d", errno, getsid (0));
	  pop3_syslog (LOG_ERR, logbuf, FALSE);
	}
  /* close inherited file descriptors */
  close (0);
  close (1);
  if ((debug & 0x100) != 0x100)
	close (2);					/* errors need to go syslog now */

  sigaction (SIGCHLD, &signalaction, NULL);
}

/* The main part of the daemon. This function reads input from the client and
   executes the proper functions. Also handles the bulk of error reporting. */

int
pop3_mainloop (int infile, int outfile, char **argv)
{
  int status = OK;
  char *buf, *arg, *cmd;
  char *local_hostname;
  struct sockaddr_storage peername;
  socklen_t namelen = sizeof (peername);
  struct timeval time_start, time_finish;
  int argv0size = strlen (argv[0]);
  union sigval sval;
  struct addrinfo hints, *gaires;
  char portstr[16];
  int ret;
  in_port_t peerport = 0;

  mailbox = NULL;
  username = NULL;
  ifile = infile;
  /* fcntl(ifile, F_SETFL, O_NONBLOCK); */
  ofile = fdopen (outfile, "w+");
  if (ofile == NULL)
	pop3_abquit (ERR_NO_OFILE);
  state = AUTHORIZATION;
  lockfile = NULL;
  curr_time = time (NULL);

  getpeername (infile, (struct sockaddr *) &peername, &namelen);
  if (peername.ss_family == AF_INET)
	{
	  inet_ntop (peername.ss_family, &((struct sockaddr_in *) &peername)->sin_addr.s_addr, peeraddr, sizeof (peeraddr));
	  peerport = ((struct sockaddr_in *) &peername)->sin_port;
	}
#ifdef USE_IPV6
  else if (peername.ss_family == AF_INET6)
	{
	  inet_ntop (peername.ss_family, &((struct sockaddr_in6 *) &peername)->sin6_addr, peeraddr, sizeof (peeraddr));
	  peerport = ((struct sockaddr_in6 *) &peername)->sin6_port;
	}
#endif
  syslog (LOG_INFO, "Incoming connection opened from %s port %d", peeraddr, peerport);
  total_sessions++;

  /* Prepare the shared secret for APOP */
  local_hostname = malloc (MAXHOSTNAMELEN + 1);
  if (local_hostname == NULL)
	pop3_abquit (ERR_NO_MEM);

  gethostname (local_hostname, MAXHOSTNAMELEN);
  memset ((char *) &hints, 0, sizeof (hints));
  hints.ai_family = (ipv6enabled ? AF_IN : AF_INET);	/* Allow IPv4 or IPv6 */
  hints.ai_socktype = SOCK_STREAM;	/* Stream socket */
  hints.ai_flags = AI_PASSIVE | AI_NUMERICSERV | AI_CANONNAME | AI_CANONIDN;	/* For wildcard IP address on specified port */
  hints.ai_protocol = (getprotobyname ("tcp"))->p_proto;	/* TCP only */
  hints.ai_canonname = NULL;
  hints.ai_addr = NULL;
  hints.ai_next = NULL;
  sprintf (portstr, "%d", port);
  ret = getaddrinfo (NULL, portstr, &hints, &gaires);
  if (ret == 0)
	{

	  /* getaddrinfo() returns a list of address structures.
	     Try each address until we successfully bind(2).
	     If socket(2) (or bind(2)) fails, we (close the socket
	     and) try the next address. */
	  free (local_hostname);
	  local_hostname = strdup (gaires->ai_canonname);
	}

  md5shared = malloc (strlen (local_hostname) + 51);
  if (md5shared == NULL)
	pop3_abquit (ERR_NO_MEM);

  snprintf (md5shared, strlen (local_hostname) + 50, "<%d.%ld@%s>", getpid (), time (NULL), local_hostname);
  free (local_hostname);

  fflush (ofile);
  fprintf (ofile, "+OK POP3 " WELCOME " %s\r\n", md5shared);

/* gnu-pop3d can run as a different user; for example, you can use
 * inetd's user field setting to run it as a different user.
 */
  if (debug & 0x04)
	syslog (LOG_INFO, "initial uid %d, gid %d", getuid (), getgid ());

  while (state != UPDATE)
	{
	  pop3_syslog (LOG_INFO, NULL, TRUE);
	  get_now (&time_start, 0);
	  fflush (ofile);
	  status = OK;
	  buf = pop3_readline (ifile);
	  cmd = pop3_cmd (buf);
	  arg = pop3_args (buf);

	  if ((debug & 0x08) && (strncasecmp (cmd, "PASS", 4) != 0))
		syslog (LOG_INFO, "cmd: %s", buf);
	  free (buf);
	  buf = NULL;
	  if (exit_flag != 0)
		{
		  status = ERR_WRONG_STATE;
		  pop3_abquit ((exit_flag >> 2 == SIGPIPE ? ERR_DEAD_SOCK : ERR_SIGNAL));
		}
	  else if (strlen (arg) + strlen (cmd) >= POP_MAXCMDLEN)
		status = ERR_TOO_LONG;
	  else if (strlen (cmd) > 4)
		status = ERR_BAD_CMD;
	  else if (strncasecmp (cmd, "USER", 4) == 0)
		{
		  strncpy (argv[0], "gnu-pop3d ", argv0size);
		  strncat (argv[0], arg, argv0size - 10);
		  status = pop3_user (arg);
		}
	  else if (strncasecmp (cmd, "QUIT", 4) == 0)
		status = pop3_quit (arg);
	  else if (strncasecmp (cmd, "APOP", 4) == 0)
		status = pop3_apop (arg);
	  else if (strncasecmp (cmd, "AUTH", 4) == 0)
		status = pop3_auth (arg);
	  else if (strncasecmp (cmd, "STAT", 4) == 0)
		status = pop3_stat (arg);
	  else if (strncasecmp (cmd, "LIST", 4) == 0)
		status = pop3_list (arg);
	  else if (strncasecmp (cmd, "RETR", 4) == 0)
		status = pop3_retr (arg);
	  else if (strncasecmp (cmd, "DELE", 4) == 0)
		status = pop3_dele (arg);
	  else if (strncasecmp (cmd, "NOOP", 4) == 0)
		status = pop3_noop (arg);
	  else if (strncasecmp (cmd, "RSET", 4) == 0)
		status = pop3_rset (arg);
	  else if ((strncasecmp (cmd, "TOP", 3) == 0) && (strlen (cmd) == 3))
		status = pop3_top (arg);
	  else if (strncasecmp (cmd, "UIDL", 4) == 0)
		status = pop3_uidl (arg);
	  else if (strncasecmp (cmd, "CAPA", 4) == 0)
		status = pop3_capa (arg);
	  else
		status = ERR_BAD_CMD;

	  if (status == OK)
		fflush (ofile);
	  else if (status == ERR_WRONG_STATE)
		fprintf (ofile, "-ERR " BAD_STATE "\r\n");
	  else if (status == ERR_BAD_ARGS)
		fprintf (ofile, "-ERR " BAD_ARGS "\r\n");
	  else if (status == ERR_NO_MESG)
		fprintf (ofile, "-ERR " NO_MESG "\r\n");
	  else if (status == ERR_NOT_IMPL)
		fprintf (ofile, "-ERR " NOT_IMPL "\r\n");
	  else if (status == ERR_BAD_CMD)
		fprintf (ofile, "-ERR " BAD_COMMAND "\r\n");
	  else if (status == ERR_BAD_LOGIN)
		{
		  sleep (1);
		  fprintf (ofile, "-ERR " BAD_LOGIN "\r\n");
		}
	  else if (status == ERR_MBOX_LOCK)
		fprintf (ofile, "-ERR [IN-USE] " MBOX_LOCK "\r\n");
	  else if (status == ERR_TOO_LONG)
		fprintf (ofile, "-ERR " TOO_LONG "\r\n");
	  else if (status == ERR_LOCK_FAILED)
		fprintf (ofile, "-ERR " LOCK_FAILED "\r\n");
	  else if (status == ERR_FILEPOS)
		fprintf (ofile, "-ERR " FILEPOS "\r\n");

	  if (status != OK)
		{
		  if (username != NULL)
			if ((buf = malloc ((strlen (username) + 66) * sizeof (char))) != NULL)
			  {
				buf = strcpy (buf, username);
#ifdef USE_VIRTUAL
				if (virtualdomain && (strchr (username, '@') == NULL))
				  {
					buf = strcat (buf, "@");
					buf = strncat (buf, virtualdomain, 64 * sizeof (char));
				  }
#endif
				syslog (LOG_ERR, "User %s cmd %s status = %d exit_flag = %ld\n", buf, cmd, status, SAFMT exit_flag);
			  }
		}

	  if (buf != NULL)
		free (buf);
	  free (cmd);
	  free (arg);
	  get_now (&time_finish, 0);
	  time_used = time_used + time_to_seconds (&time_start, &time_finish);
	  time_calls++;
	}

#ifdef USE_VIRTUAL
  if (virtualdomain)
	{
	  free (virtualdomain);
	  virtualdomain = NULL;
	}
#endif /* USE_VIRTUAL */
  fflush (ofile);
  if (time_used > 0.0)
	syslog (LOG_INFO, "Time used: %.2f sec in %i commands by %s from %s", time_used, time_calls, username, peeraddr);
  if (username)
	{
	  free (username);
	  username = NULL;
	}
  sval.sival_int = 0;
  if (daemonpid != getpid ())
	sigqueue (daemonpid, SIGUSR1, sval);
  return OK;
}

/* Runs GNU POP in standalone daemon mode. This opens and binds to a port
   (default 110) then executes a pop3_mainloop() upon accepting a connection.
   It starts maxchildren child processes to listen to and accept socket
   connections */

int
pop3_daemon (int maxchildren, char **argv)
{
  struct sockaddr_storage client;
  struct addrinfo hints, *gaires, *rp;
  char portstr[16];
  int sock2;
  unsigned int socksize = sizeof (client);
  int flag = 1;
  int ret;
  pid_t pid;
  sigset_t blockset;
  union sigval sval;
  char logbuf[128];

  sigemptyset (&blockset);
  sigaddset (&blockset, SIGUSR1);
  sigaddset (&blockset, SIGCHLD);

  if (debug & 0x01)
	{
	  sprintf (logbuf, "pop3_daemon: starting on port %d", port);
	  pop3_syslog (LOG_INFO, logbuf, FALSE);
	}

  memset ((char *) &hints, 0, sizeof (hints));
  hints.ai_family = (ipv6enabled ? AF_IN : AF_INET);	/* Allow IPv4 or IPv6 */
  hints.ai_socktype = SOCK_STREAM;	/* Stream socket */
  hints.ai_flags = AI_PASSIVE | AI_NUMERICSERV;	/* For wildcard IP address on specified port */
  hints.ai_protocol = (getprotobyname ("tcp"))->p_proto;	/* TCP only */
  hints.ai_canonname = NULL;
  hints.ai_addr = NULL;
  hints.ai_next = NULL;
  sprintf (portstr, "%d", port);
  ret = getaddrinfo (NULL, portstr, &hints, &gaires);
  if (ret != 0)
	{
	  syslog (LOG_ERR, "pop3_daemon: getaddrinfo: %s", gai_strerror (ret));
	  exit (1);
	}

  /* getaddrinfo() returns a list of address structures.
     Try each address until we successfully bind(2).
     If socket(2) (or bind(2)) fails, we (close the socket
     and) try the next address. */

  for (rp = gaires; rp != NULL; rp = rp->ai_next)
	{
	  sock = socket (rp->ai_family, rp->ai_socktype, rp->ai_protocol);
	  if (sock < 0)
		continue;

	  flag = 1;
	  ret = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof (flag));	/* not having to wait for reusing the port */
	  if (ret != 0)
		syslog (LOG_ERR, "pop3_daemon: %s error=%d", "Could not setsockopt(SO_REUSEADDR)", ret);
#ifdef USE_IPV6
	  flag = 0;
	  ret = 0;
	  if (ipv6enabled == 1)
		ret = setsockopt (sock, IPPROTO_IPV6, IPV6_V6ONLY, (char *) &flag, sizeof (flag));	/* not having to wait for reusing the port */
	  if (ret != 0)
		syslog (LOG_ERR, "pop3_daemon: Could not setsockopt(IPPROTO_IPV6, IPV6_V6ONLY) = %d error=%d", flag, ret);
#endif

	  if (bind (sock, rp->ai_addr, rp->ai_addrlen) == 0)
		break;					/* Success */
	  close (sock);
	}
  if (rp == NULL)
	{							/* No address succeeded */
	  syslog (LOG_ERR, "pop3_daemon:Couldn't bind to socket");
	  perror ("pop3_daemon:Couldn't bind to socket");
	  exit (1);
	}

  freeaddrinfo (gaires);		/* No longer needed */

  listen (sock, 128);

  while ((exit_flag == 0) || ((exit_flag == 1) && (child_procs > 0)))
	{
	  pop3_syslog (LOG_INFO, NULL, TRUE);
	  sigprocmask (SIG_UNBLOCK, &blockset, NULL);

	  if ((child_procs < maxchildren) && (exit_flag == 0))
		{
		  if (debug & 0x04)
			{
			  sprintf (logbuf, "pop3_daemon: waiting (accept) for new client, children: %ld", SAFMT child_procs);
			  pop3_syslog (LOG_INFO, logbuf, TRUE);
			}

		  if ((sock2 = accept (sock, (struct sockaddr *) &client, &socksize)) <= -1)
			{
			  if (debug & 0x04)
				{
				  sprintf (logbuf, "pop3_daemon: accept new client interrupted errno: %d, children: %ld exit_flag: %ld", errno, SAFMT child_procs, SAFMT exit_flag);	/* EINTR = 4 */
				  pop3_syslog (LOG_INFO, logbuf, FALSE);
				}
			  continue;
			}
		  else
			pop3_syslog (LOG_INFO, NULL, TRUE);

		  if (debug & 0x04)
			{
			  sprintf (logbuf, "pop3_daemon: accepted new client, children: %ld exit_flag: %ld", SAFMT child_procs,
					   SAFMT exit_flag);
			  pop3_syslog (LOG_INFO, logbuf, TRUE);
			}

		  sigprocmask (SIG_BLOCK, &blockset, NULL);
		  if ((pid = fork ()) < 0)
			{					/* can't fork */
			  sprintf (logbuf, "Can't start server %ld", SAFMT child_procs);
			  pop3_syslog (LOG_ERR, logbuf, TRUE);
			}
		  else if (pid == 0)
			{					/* following code executes in child process */
			  last_signal = 0;
			  sigprocmask (SIG_UNBLOCK, &blockset, NULL);
			  if (debug & 0x40)
				{
				  sprintf (logbuf, "Session leader: %d", getsid (0));
				  syslog (LOG_INFO, "%s", logbuf);
				}
			  close (sock);
			  sock = sock2;
			  /* Disable the Nagle (TCP No Delay) algorithm */
			  ret = setsockopt (sock2, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof (flag));
			  if (ret != 0)
				{
				  sprintf (logbuf, "%s error=%d", "Could not setsockopt(TCP_NODELAY)", ret);
				  syslog (LOG_ERR, "%s", logbuf);
				}

			  pop3_mainloop (sock2, sock2, argv);
			  shutdown (sock2, SHUT_RDWR);	/* SHUT_RDWR = 2 */
			  close (sock2);
			  _exit (OK);
			}
		  else
			{					/* main process */
			  total_sessions++;
			  sval.sival_int = 0;
			  if (sigqueue (pid, 0, sval) == 0)	/* child not immediately terminated */
				ret = pop3_childpidsupdate (1, pid);
			  if (debug & 0x01)
				{
				  sprintf (logbuf, "pop3_daemon: child %ld pid %d started exit_flag: %ld", SAFMT child_procs, pid,
						   SAFMT exit_flag);
				  pop3_syslog (LOG_INFO, logbuf, FALSE);
				}
			  close (sock2);
			}
		}
	  else
		{
		  sprintf (logbuf, "pop3_daemon: children limit reached, pausing, children: %ld exit_flag: %ld",
				   SAFMT child_procs, SAFMT exit_flag);
		  pop3_syslog (LOG_WARNING, logbuf, TRUE);
		  if ((child_procs >= maxchildren) || ((exit_flag == 1) && (child_procs > 0)))
			pause ();
		  sprintf (logbuf, "pop3_daemon: resuming operation after pause, children: %ld exit_flag: %ld",
				   SAFMT child_procs, SAFMT exit_flag);
		  pop3_syslog (LOG_WARNING, logbuf, FALSE);
		}
	}
  return OK;
}
