#!/usr/bin/perl
# mkpwd.pl [password|length >= 3]
# create and/or encrypt a password for a virtual domain
# using the same original DES algorithm as Linux crypt
#
use strict;
use warnings;
use Crypt::Password;
 
my $salt = join '', ('.', '/', 0..9, 'A'..'Z', 'a'..'z')[rand 64, rand 64];
my $password = $ARGV[0] || 8;

if ($password =~ /^-h|--help/) {
	print "Generation and/or hashing of password compatible with glibc crypt()\n";
	print "$0 <length of generated cleartext password>\n";
	print "$0 <cleartext password to be hashed>\n";
exit;
} 
if ($password =~ /^[0-9]{1,2}$/) {  # password without cipher zero, one and letter l, O, with special char and both upper and lower case char
   my $i = $password-4; 
   $password = ('a'..'k', 'm'..'z')[rand 25];
   $password .= ('*','#','!','-','+','=','(',')')[rand 8];
   $password .= ('2'..'9')[rand 8];
   $password .= ('A'..'N','P'..'Z')[rand 26];
	
   $password .= join '', ( map { ('*','#','!','-','+','=','(',')','a'..'k','m'..'z','A'..'N','P'..'Z','2'..'9')[rand 64] } 1..$i) if $i > 0;
}

sub sha {
my ($password, $salt, $id) = @_;
my $encrypted = '';
$encrypted = password($password, $salt, 'sha256') if $id == 5;
$encrypted = password($password, $salt, 'sha512') if $id == 6;
return $encrypted;
}

sub des {
my ($password, $salt) = @_;
my $encrypted = crypt($password, $salt);
return $encrypted;
}

my $encryptedDES = des($password, $salt);
my $encryptedSHA256 = sha($password, $salt, 5);
my $encryptedSHA512 = sha($password, $salt, 6);

print "Password (".length($password)."): $password  (using salt: $salt)\n";
print "\nDES (original crypt) salt (2) + encrypted (11) (to be stored):\n $encryptedDES\n (DO NOT USE, only first 8 chars of password significant, stolen encrypted password can login)\n";
print "\nSHA256 '\$5\$salt\$' + encrypted (".(length($encryptedSHA256)-6).") (to be stored):\n $encryptedSHA256\n";
print "\nSHA512 '\$6\$salt\$' + encrypted (".(length($encryptedSHA512)-6).") (to be stored):\n $encryptedSHA512\n";

# In /etc/virtual/<domain>/passwd
#
# kristian:password18:1511:100:Kristian V. Thomsen:.: 
# kristian:YuBTPqxwN/o/k:1511:100:Kristian V. Thomsen:.: 
